require('./utils');
const lex = require('../lib/nql').lex;

describe('Lexer', function () {
    const lexicalError = /^Query Error: unrecognized text/;

    describe('Symbols', function () {
        it('can recognise -', function () {
            lex('-').should.eql([{token: 'NOT', matched: '-'}]);
        });
        it('can recognise +', function () {
            lex('+').should.eql([{token: 'AND', matched: '+'}]);
        });
        it('can recognise ,', function () {
            lex(',').should.eql([{token: 'OR', matched: ','}]);
        });
        it('can recognise [', function () {
            lex('[').should.eql([{token: 'LBRACKET', matched: '['}]);
        });
        it('can recognise ]', function () {
            lex(']').should.eql([{token: 'RBRACKET', matched: ']'}]);
        });
        it('can recognise (', function () {
            lex('(').should.eql([{token: 'LPAREN', matched: '('}]);
        });
        it('can recognise )', function () {
            lex(')').should.eql([{token: 'RPAREN', matched: ')'}]);
        });
        it('can recognise >', function () {
            lex('>').should.eql([{token: 'GT', matched: '>'}]);
        });
        it('can recognise <', function () {
            lex('<').should.eql([{token: 'LT', matched: '<'}]);
        });
        it('can recognise >=', function () {
            lex('>=').should.eql([{token: 'GTE', matched: '>='}]);
        });
        it('can recognise <=', function () {
            lex('<=').should.eql([{token: 'LTE', matched: '<='}]);
        });
        it('cannot recognise :', function () {
            (function () {
                lex(':');
            }).should.throw(lexicalError);
        });
        it('cannot recognise =', function () {
            (function () {
                lex('=');
            }).should.throw(lexicalError);
        });
        it('cannot recognise "', function () {
            (function () {
                lex('"');
            }).should.throw(lexicalError);
        });
        it('cannot recognise \'', function () {
            (function () {
                lex('\'');
            }).should.throw(lexicalError);
        });
    });

    describe('VALUES', function () {
        it('can recognise null', function () {
            lex('null').should.eql([{token: 'NULL', matched: 'null'}]);
        });

        it('can recognise true', function () {
            lex('true').should.eql([{token: 'TRUE', matched: 'true'}]);
        });

        it('can recognise false', function () {
            lex('false').should.eql([{token: 'FALSE', matched: 'false'}]);
        });

        it('can recognise a LITERAL', function () {
            lex('six').should.eql([{token: 'LITERAL', matched: 'six'}]);
        });

        it('can recognise a STRING', function () {
            lex('\'six\'').should.eql([{token: 'STRING', matched: '\'six\''}]);
        });

        it('can recognise a NUMBER', function () {
            lex('6').should.eql([{token: 'NUMBER', matched: '6'}]);
        });

        it('does not confuse values in LITERALs', function () {
            lex('strueth').should.eql([{token: 'LITERAL', matched: 'strueth'}]);
            lex('trueth').should.eql([{token: 'LITERAL', matched: 'trueth'}]);
            lex('true_thing').should.eql([{token: 'LITERAL', matched: 'true_thing'}]);
            // lex("true-thing").should.eql([{token: "LITERAL", matched: "true-thing"}]);
        });

        it('does not confuse values in STRINGs', function () {
            lex('\'strueth\'').should.eql([{token: 'STRING', matched: '\'strueth\''}]);
            lex('\'trueth\'').should.eql([{token: 'STRING', matched: '\'trueth\''}]);
            lex('\'true_thing\'').should.eql([{token: 'STRING', matched: '\'true_thing\''}]);
            lex('\'true-thing\'').should.eql([{token: 'STRING', matched: '\'true-thing\''}]);
        });
    });

    describe('LITERAL values', function () {
        it('should match literals', function () {
            lex('myvalue').should.eql([
                {token: 'LITERAL', matched: 'myvalue'}
            ]);
            lex('my value').should.eql([
                {token: 'LITERAL', matched: 'my'},
                {token: 'LITERAL', matched: 'value'}
            ]);
            lex('my-value').should.eql([
                {token: 'LITERAL', matched: 'my-value'}
            ]);
            lex('my&value!').should.eql([
                {token: 'LITERAL', matched: 'my&value!'}
            ]);
            lex('my&valu\\\'e!').should.eql([
                {token: 'LITERAL', matched: 'my&valu\\\'e!'}
            ]);
            (function () {
                lex('my&valu\'e!');
            }).should.throw(lexicalError);
        });

        it('should separate NOT at beginning of literal', function () {
            lex('-photo').should.eql([
                {token: 'NOT', matched: '-'},
                {token: 'LITERAL', matched: 'photo'}
            ]);

            lex('-photo-graph').should.eql([
                {token: 'NOT', matched: '-'},
                {token: 'LITERAL', matched: 'photo-graph'}
            ]);
        });

        it('should NOT permit special chars inside a literal', function () {
            (function () {
                lex('t+st');
            }).should.throw(lexicalError);
            (function () {
                lex('t,st');
            }).should.throw(lexicalError);
            (function () {
                lex('t(st');
            }).should.throw(lexicalError);
            (function () {
                lex('t)st');
            }).should.throw(lexicalError);
            (function () {
                lex('t>st');
            }).should.throw(lexicalError);
            (function () {
                lex('t<st');
            }).should.throw(lexicalError);
            (function () {
                lex('t=st');
            }).should.throw(lexicalError);
            (function () {
                lex('t[st');
            }).should.throw(lexicalError);
            (function () {
                lex('t]st');
            }).should.throw(lexicalError);
            (function () {
                lex('t\'st');
            }).should.throw(lexicalError);
            (function () {
                lex('t"st');
            }).should.throw(lexicalError);
        });

        it('should not match special chars at the start of a literal', function () {
            lex('+test').should.eql([
                {token: 'AND', matched: '+'},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex(',test').should.eql([
                {token: 'OR', matched: ','},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex('(test').should.eql([
                {token: 'LPAREN', matched: '('},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex(')test').should.eql([
                {token: 'RPAREN', matched: ')'},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex('>test').should.eql([
                {token: 'GT', matched: '>'},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex('<test').should.eql([
                {token: 'LT', matched: '<'},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex('[test').should.eql([
                {token: 'LBRACKET', matched: '['},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex(']test').should.eql([
                {token: 'RBRACKET', matched: ']'},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex('>=test').should.eql([
                {token: 'GTE', matched: '>='},
                {token: 'LITERAL', matched: 'test'}
            ]);
            lex('<=test').should.eql([
                {token: 'LTE', matched: '<='},
                {token: 'LITERAL', matched: 'test'}
            ]);

            (function () {
                lex('=test');
            }).should.throw(lexicalError);
            (function () {
                lex('"test');
            }).should.throw(lexicalError);
            (function () {
                lex('\'test');
            }).should.throw(lexicalError);
        });

        it('should not match special chars at the end of a literal', function () {
            lex('test+').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'AND', matched: '+'}
            ]);
            lex('test,').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'OR', matched: ','}
            ]);
            lex('test(').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'LPAREN', matched: '('}
            ]);
            lex('test)').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'RPAREN', matched: ')'}
            ]);
            lex('test>').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'GT', matched: '>'}
            ]);
            lex('test<').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'LT', matched: '<'}
            ]);
            lex('test[').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'LBRACKET', matched: '['}
            ]);
            lex('test]').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'RBRACKET', matched: ']'}
            ]);
            lex('test>=').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'GTE', matched: '>='}
            ]);
            lex('test<=').should.eql([
                {token: 'LITERAL', matched: 'test'},
                {token: 'LTE', matched: '<='}
            ]);
            (function () {
                lex('test=');
            }).should.throw(lexicalError);
            (function () {
                lex('test"');
            }).should.throw(lexicalError);
            (function () {
                lex('test\'');
            }).should.throw(lexicalError);
        });

        it('should permit escaped special chars inside a literal', function () {
            lex('t\\+st').should.eql([{token: 'LITERAL', matched: 't\\+st'}]);
            lex('t\\,st').should.eql([{token: 'LITERAL', matched: 't\\,st'}]);
            lex('t\\(st').should.eql([{token: 'LITERAL', matched: 't\\(st'}]);
            lex('t\\)st').should.eql([{token: 'LITERAL', matched: 't\\)st'}]);
            lex('t\\>st').should.eql([{token: 'LITERAL', matched: 't\\>st'}]);
            lex('t\\<st').should.eql([{token: 'LITERAL', matched: 't\\<st'}]);
            lex('t\\=st').should.eql([{token: 'LITERAL', matched: 't\\=st'}]);
            lex('t\\[st').should.eql([{token: 'LITERAL', matched: 't\\[st'}]);
            lex('t\\]st').should.eql([{token: 'LITERAL', matched: 't\\]st'}]);
            lex('t\\\'st').should.eql([{token: 'LITERAL', matched: 't\\\'st'}]);
            lex('t\\"st').should.eql([{token: 'LITERAL', matched: 't\\"st'}]);
        });
    });

    describe('LITERAL vs PROP', function () {
        it('should match colon in string as PROP before, literal after', function () {
            lex(':test').should.eql([
                {token: 'LITERAL', matched: ':test'}
            ]);

            lex('te:st').should.eql([
                {token: 'PROP', matched: 'te:'},
                {token: 'LITERAL', matched: 'st'}
            ]);

            lex('test:').should.eql([
                {token: 'PROP', matched: 'test:'}
            ]);
        });

        it('should only match colon-at-end as PROP if PROP is valPROP', function () {
            lex('te!:st').should.eql([
                {token: 'LITERAL', matched: 'te!:st'}
            ]);

            lex('post-count:6').should.eql([
                {token: 'LITERAL', matched: 'post-count:6'}
            ]);

            lex('post_count:6').should.eql([
                {token: 'PROP', matched: 'post_count:'},
                {token: 'NUMBER', matched: '6'}
            ]);
        });
    });

    describe('STRING values', function () {
        it('can recognise simple STRING', function () {
            lex('\'magic\'').should.eql([{token: 'STRING', matched: '\'magic\''}]);
            lex('\'magic mystery\'').should.eql([{token: 'STRING', matched: '\'magic mystery\''}]);
            lex('\'magic 123\'').should.eql([{token: 'STRING', matched: '\'magic 123\''}]);
        });

        it('can recognise multiple STRING values', function () {
            lex('\'magic\'\'mystery\'').should.eql([
                {token: 'STRING', matched: '\'magic\''},
                {token: 'STRING', matched: '\'mystery\''}
            ]);
            lex('\'magic\' \'mystery\'').should.eql([
                {token: 'STRING', matched: '\'magic\''},
                {token: 'STRING', matched: '\'mystery\''}
            ]);
            lex('\'magic\',\'mystery\'').should.eql([
                {token: 'STRING', matched: '\'magic\''},
                {token: 'OR', matched: ','},
                {token: 'STRING', matched: '\'mystery\''}
            ]);
            lex('[\'magic\',\'mystery\']').should.eql([
                {token: 'LBRACKET', matched: '['},
                {token: 'STRING', matched: '\'magic\''},
                {token: 'OR', matched: ','},
                {token: 'STRING', matched: '\'mystery\''},
                {token: 'RBRACKET', matched: ']'}
            ]);
        });

        it('can recognise STRING with special characters', function () {
            lex('\'magic+\'').should.eql([{token: 'STRING', matched: '\'magic+\''}]);
            lex('\'magic,\'').should.eql([{token: 'STRING', matched: '\'magic,\''}]);
            lex('\'magic-\'').should.eql([{token: 'STRING', matched: '\'magic-\''}]);
            lex('\'magic>\'').should.eql([{token: 'STRING', matched: '\'magic>\''}]);
            lex('\'magic<\'').should.eql([{token: 'STRING', matched: '\'magic<\''}]);
        });

        it('should permit special chars inside a STRING, not including quotes', function () {
            lex('\'t+st\'').should.eql([{token: 'STRING', matched: '\'t+st\''}]);
            lex('\'t,st\'').should.eql([{token: 'STRING', matched: '\'t,st\''}]);
            lex('\'t(st\'').should.eql([{token: 'STRING', matched: '\'t(st\''}]);
            lex('\'t)st\'').should.eql([{token: 'STRING', matched: '\'t)st\''}]);
            lex('\'t>st\'').should.eql([{token: 'STRING', matched: '\'t>st\''}]);
            lex('\'t<st\'').should.eql([{token: 'STRING', matched: '\'t<st\''}]);
            lex('\'t=st\'').should.eql([{token: 'STRING', matched: '\'t=st\''}]);
            lex('\'t[st\'').should.eql([{token: 'STRING', matched: '\'t[st\''}]);
            lex('\'t]st\'').should.eql([{token: 'STRING', matched: '\'t]st\''}]);
        });

        it('should NOT permit quotes inside a STRING', function () {
            (function () {
                lex('\'t\'st\'');
            }).should.throw(lexicalError);
            (function () {
                lex('\'t"st\'');
            }).should.throw(lexicalError);
        });

        it('should permit escaped quotes inside a String', function () {
            lex('\'t\\\'st\'').should.eql([{token: 'STRING', matched: '\'t\\\'st\''}]);
            lex('\'t\\"st\'').should.eql([{token: 'STRING', matched: '\'t\\"st\''}]);
        });
    });

    describe('single & double QUOTE marks', function () {
        it('CANNOT match an UNescaped double quote in a LITERAL', function () {
            (function () {
                lex('thing"amabob');
            }).should.throw(lexicalError);
        });
        it('CANNOT match an UNescaped single quote in a LITERAL', function () {
            (function () {
                lex('thing\'amabob');
            }).should.throw(lexicalError);
        });
        it('CANNOT match an UNescaped double quote in a STRING', function () {
            (function () {
                lex('\'thing"amabob\'');
            }).should.throw(lexicalError);
        });
        it('CANNOT match an UNescaped single quote in a STRING', function () {
            (function () {
                lex('\'thing\'amabob\'');
            }).should.throw(lexicalError);
        });
        it('CAN match an escaped double quote in a LITERAL', function () {
            lex('thing\\"amabob').should.eql([{token: 'LITERAL', matched: 'thing\\"amabob'}]);
        });
        it('CAN match an escaped single quote in a LITERAL', function () {
            lex('thing\\\'amabob').should.eql([{token: 'LITERAL', matched: 'thing\\\'amabob'}]);
        });
        it('CAN match an escaped double quote in a STRING', function () {
            lex('\'thing\\"amabob\'').should.eql([{token: 'STRING', matched: '\'thing\\"amabob\''}]);
        });
        it('CAN match an escaped single quote in a STRING', function () {
            lex('\'thing\\\'amabob\'').should.eql([{token: 'STRING', matched: '\'thing\\\'amabob\''}]);
        });
    });

    describe('Filter expressions', function () {
        it('should separate NOT at beginning of literal', function () {
            lex('tag:-photo').should.eql([
                {token: 'PROP', matched: 'tag:'},
                {token: 'NOT', matched: '-'},
                {token: 'LITERAL', matched: 'photo'}
            ]);

            lex('tag:-photo-graph').should.eql([
                {token: 'PROP', matched: 'tag:'},
                {token: 'NOT', matched: '-'},
                {token: 'LITERAL', matched: 'photo-graph'}
            ]);

            lex('tags:[-getting-started]').should.eql([
                {token: 'PROP', matched: 'tags:'},
                {token: 'LBRACKET', matched: '['},
                {token: 'NOT', matched: '-'},
                {token: 'LITERAL', matched: 'getting-started'},
                {token: 'RBRACKET', matched: ']'}
            ]);
        });

        it('should permit NOT inside a literal', function () {
            lex('tags:getting-started').should.eql([
                {token: 'PROP', matched: 'tags:'},
                {token: 'LITERAL', matched: 'getting-started'}
            ]);

            lex('tags:[getting-started]').should.eql([
                {token: 'PROP', matched: 'tags:'},
                {token: 'LBRACKET', matched: '['},
                {token: 'LITERAL', matched: 'getting-started'},
                {token: 'RBRACKET', matched: ']'}
            ]);

            lex('tags:-[getting-started]').should.eql([
                {token: 'PROP', matched: 'tags:'},
                {token: 'NOT', matched: '-'},
                {token: 'LBRACKET', matched: '['},
                {token: 'LITERAL', matched: 'getting-started'},
                {token: 'RBRACKET', matched: ']'}
            ]);

            lex('id:-1+tags:[getting-started]').should.eql([
                {token: 'PROP', matched: 'id:'},
                {token: 'NOT', matched: '-'},
                {token: 'NUMBER', matched: '1'},
                {token: 'AND', matched: '+'},
                {token: 'PROP', matched: 'tags:'},
                {token: 'LBRACKET', matched: '['},
                {token: 'LITERAL', matched: 'getting-started'},
                {token: 'RBRACKET', matched: ']'}
            ]);
        });
    });

    describe('complex examples', function () {
        it('many expressions', function () {
            lex('tag:photo+featured:true,tag.count:>5').should.eql([
                {token: 'PROP', matched: 'tag:'},
                {token: 'LITERAL', matched: 'photo'},
                {token: 'AND', matched: '+'},
                {token: 'PROP', matched: 'featured:'},
                {token: 'TRUE', matched: 'true'},
                {token: 'OR', matched: ','},
                {token: 'PROP', matched: 'tag.count:'},
                {token: 'GT', matched: '>'},
                {token: 'NUMBER', matched: '5'}
            ]);

            // lex("tag:photo+image:-null,tag.count:>5").should.eql();
        });

        it('grouped expressions', function () {
            // lex("author:-joe+(tag:photo,image:-null,featured:true)").should.eql();
        });

        it('in expressions', function () {
            lex('author:-joe+tag:[photo,video]').should.eql([
                {token: 'PROP', matched: 'author:'},
                {token: 'NOT', matched: '-'},
                {token: 'LITERAL', matched: 'joe'},
                {token: 'AND', matched: '+'},
                {token: 'PROP', matched: 'tag:'},
                {token: 'LBRACKET', matched: '['},
                {token: 'LITERAL', matched: 'photo'},
                {token: 'OR', matched: ','},
                {token: 'LITERAL', matched: 'video'},
                {token: 'RBRACKET', matched: ']'}
            ]);

            // lex("author:-joe+tag:-[photo,video]").should.eql();

            // lex("author:-joe+tag:[photo,video]+post.count:>5+post.count:<100").should.eql();
        });
    });
});
